/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

// Testing featured-based get_experiment_branch semantics.

#[cfg(feature = "rkv-safe-mode")]
mod common;

#[allow(unused_imports)]
use nimbus::error::Result;

#[cfg(feature = "rkv-safe-mode")]
#[test]
fn test_feature_before_open() -> Result<()> {
    use nimbus::error::NimbusError;

    let _ = env_logger::try_init();
    let client = common::new_test_client("test_feature_before_open")?;
    assert!(matches!(
        client.get_experiment_branch("not_there_feature".to_string()),
        Err(NimbusError::DatabaseNotReady)
    ));
    // now initialize the DB - it should start working (and report no branch)
    client.initialize()?;
    assert_eq!(
        client.get_experiment_branch("not_there_feature".to_string())?,
        None
    );
    Ok(())
}

#[cfg(feature = "rkv-safe-mode")]
#[test]
fn test_enrolled_feature() -> Result<()> {
    let _ = env_logger::try_init();
    let client = common::new_test_client("test_enrolled_feature")?;
    client.initialize()?;
    client.set_experiments_locally(common::experiments_testing_feature_ids())?;

    // haven't applied them yet, so not enrolled as the experiment doesn't
    // really exist.
    assert_eq!(
        client.get_experiment_branch("aboutwelcome".to_string())?,
        None,
        "shouldn't return anything before pending experiments applied"
    );

    client.apply_pending_experiments()?;

    // aboutwelcome enrolls everyone - not clear what treatment though.
    assert!(
        client
            .get_experiment_branch("aboutwelcome".to_string())?
            .is_some(),
        "should return a branch for an exp that enrolls everyone"
    );

    client.opt_out("secure-gold".to_string())?;
    assert_eq!(
        client.get_experiment_branch("aboutwelcome".to_string())?,
        None,
        "should not return a branch we've just opted out of"
    );

    client.opt_in_with_branch("secure-gold".to_string(), "treatment".to_string())?;
    assert_eq!(
        client.get_experiment_branch("aboutwelcome".to_string())?,
        Some("treatment".to_string()),
        "should return a treatment branch we've just opted into"
    );

    client.opt_in_with_branch("secure-gold".to_string(), "control".to_string())?;
    assert_eq!(
        client.get_experiment_branch("aboutwelcome".to_string())?,
        Some("control".to_string()),
        "should return a second branch we've just opted into"
    );

    client.set_global_user_participation(false)?;
    assert_eq!(
        client.get_experiment_branch("aboutwelcome".to_string())?,
        None,
        "should not return a branch if we've globally opted out"
    );

    Ok(())
}
